"""
    通过线程执行函数，简化创建线程
"""
import threading
from typing import Callable


class RunByThreadDecorator:
    """
        通过线程执行函数，简化创建线程
    """

    def __init__(self, callback: Callable = None, **kwargs):
        """

        :param callback: 回调函数
        :param kwargs: 线程执行参数
        """
        self.callback = callback
        self.kwargs = kwargs

        self.func_args = None
        self.func_kwargs = None

    def run(self, func: Callable) -> None:
        """

        :param func: 执行函数
        :return:
        """
        res = func(*self.func_args, **self.func_kwargs)
        if self.callback:
            self.callback(res)

    def __call__(self, func: Callable):
        def _wrapper(*args, **kwargs):
            self.func_args = args
            self.func_kwargs = kwargs
            spider = args[0]

            thread = threading.Thread(target=self.run, args=(func,), **self.kwargs)

            if hasattr(spider, 'distribute_thread_list'):
                spider.distribute_thread_list.append(thread)  # 将线程存入列表

            thread.start()

        return _wrapper



"""
    jump spider 控制器
"""
import inspect
from palp import settings
from loguru import logger
from typing import Union, List, Callable
from palp.network.request import Request
from palp.network.response import Response
from palp.sequence.sequence import Sequence
from palp.exception import DropRequestException


class JumpController:
    REQUEST_MIDDLEWARE = {}

    def __init__(self, q: Sequence, request_middleware: Union[List[Callable], Callable], spider):
        """

        :param q: 请求队列
        :param request_middleware: 请求中间件
        :param spider: jump spider
        """
        # 导入设置
        if request_middleware is None:
            request_middleware = []
        elif not isinstance(request_middleware, list):
            request_middleware = [request_middleware]

        # 实例化
        for i, m in enumerate(request_middleware):
            if m not in self.__class__.REQUEST_MIDDLEWARE:
                request_middleware[i] = m()
                self.__class__.REQUEST_MIDDLEWARE[m] = request_middleware[i]
            else:
                request_middleware[i] = self.__class__.REQUEST_MIDDLEWARE[m]

        self.queue = q
        self.spider = spider
        self.request_middleware = request_middleware

    def run(self):
        """
        执行流程

        :return:
        """
        request = None

        while True:
            task = self.queue.get(timeout=0.5)

            try:
                if task is None:
                    break
                elif not isinstance(task, Request):
                    continue

                request = task
                self.run_requests(request=task)
            except DropRequestException as e:
                logger.warning(f"丢弃请求：{e.args}")

        if request:
            self.spider.jump_out(request)

    def add_new_request(self, new_request: Request, old_request: Request, response: Response):
        """
        添加新的请求到队列中

        :param new_request:
        :param old_request:
        :param response:
        :return:
        """
        # 自动衔接 meta
        new_request.meta = old_request.meta

        # 自动拼接 url
        if not new_request.url.startswith('http') and response:
            new_request.url = response.urljoin(new_request.url)

        # 添加请求必须参数
        new_request.cookie_jar = old_request.cookie_jar  # 续上上一个的 cookie_jar

        # 修改优先级，深层的函数应该优先处理，避免积压不前（深度爬取）
        if settings.REQUEST_QUEUE_MODE == 3:
            new_request.priority = old_request.priority - 1

        self.queue.put(new_request)

    def run_requests(self, request: Request):
        """
        处理 yield 的 Request

        :param request:
        :return:
        """
        # 请求发起前的处理
        for middleware in self.request_middleware:
            middleware.request_in(self.spider, request)

        # 处理请求
        response = None
        failed_times = 0
        while failed_times < settings.REQUEST_RETRY_TIMES:
            try:
                response = request.send()
                break
            except DropRequestException as e:
                raise DropRequestException(self.spider.name, request, *e.args)
            except Exception as e:
                failed_times += 1

                # 出现错误可直接处理，并返回新的请求，同时阻断当前错误请求传播
                for middleware in self.request_middleware:
                    new_request = middleware.request_error(self.spider, request, e)
                    if new_request is None:
                        continue
                    elif isinstance(new_request, Request):
                        self.add_new_request(new_request=new_request, old_request=request, response=response)
                        raise DropRequestException(middleware.__class__.__name__, str(request))
                    else:
                        logger.warning("request_error 仅支持 Request 返回值！")

            # 失败达到最大次数直接丢弃
            if failed_times >= settings.REQUEST_RETRY_TIMES:
                for middleware in self.request_middleware:
                    middleware.request_failed(self.spider, request)
                return

        # 请求结束的回调（可在此时判断返回值是否合理，不合理可在此重新构造请求，同时阻断当前请求传播）
        for middleware in self.request_middleware:
            new_request = middleware.request_close(self.spider, request, response)
            if new_request is None:
                continue
            elif isinstance(new_request, Request):
                self.add_new_request(new_request=new_request, old_request=request, response=response)
                raise DropRequestException(middleware.__class__.__name__, str(request))
            else:
                logger.warning("request_close 仅支持 Request 返回值！")

        # 继续监控下次 yield
        if request.callback is not None:
            if inspect.isgeneratorfunction(request.callback):  # 很好用的库，判断是否是 yield 函数
                for task in request.callback(request, response):
                    if isinstance(task, Request):
                        self.add_new_request(new_request=task, old_request=request, response=response)
            else:
                request.callback(request, response)



"""
    请求记录中间件

    目前记录：
        all：请求数
        succeed：请求成功数
        failed：请求失败数

    注意：
        通过引入 spider 自身的 spider_record 获取、修改对应信息
"""
from typing import Union
from palp.network.request import Request
from palp.decorator.decorator_lock import FuncLockDecorator
from palp.middleware.middleware_request import RequestMiddleware


class RequestRecordMiddleware(RequestMiddleware):
    """
        请求记录中间件
    """

    @FuncLockDecorator()
    def request_in(self, spider, request) -> None:
        """
            记录进入的请求
        """
        spider.spider_record['all'] += 1

    @FuncLockDecorator()
    def request_failed(self, spider, request) -> None:
        """
            记录失败的请求
        """
        spider.spider_record['failed'] += 1

    @FuncLockDecorator()
    def request_close(self, spider, request, response) -> Union[Request, None]:
        """
            记录成功的请求
        """
        spider.spider_record['succeed'] += 1

        return



"""
    通过 item 将数据传输至 pipeline
"""
import palp


class LazyItem(palp.LazyItem):
    """
        通用、懒人 item
    """


class StrictItem(palp.StrictItem):
    """
        严格 item
    """
    # 此处需要定义数据库字段
    # name = palp.Field()



"""
    redis request 队列
"""
from palp import settings
from palp.network.request import LoadRequest
from palp.sequence.sequence import RedisSequence


class FIFORequestRedisSequence(RedisSequence):
    """
        先进先出队列
    """

    @classmethod
    def get_redis_key(cls):
        """
        获取 redis 的键

        :return:
        """
        return settings.REDIS_KEY_QUEUE_REQUEST

    def put(self, obj, timeout=None):
        """
        添加任务

        :param obj:
        :param timeout:
        :return:
        """
        from palp.conn import redis_conn

        redis_conn.rpush(self.redis_key, obj.to_json())

    def get(self, timeout=None):
        """
        获取任务（这里是返回的对象）

        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.blpop(self.redis_key, timeout=timeout)
        if result:
            return LoadRequest.load_from_json(result[-1].decode())

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.llen(self.redis_key) == 0


class LIFORequestRedisSequence(FIFORequestRedisSequence):
    """
        后进先出队列
    """

    def get(self, timeout=None):
        """
        获取任务
        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.brpop(self.redis_key, timeout=timeout)
        if result:
            return LoadRequest.load_from_json(result[-1].decode())

class PriorityRequestRedisSequence(FIFORequestRedisSequence):
    """
        优先级队列
    """

    def put(self, obj, timeout=None):
        """
        添加任务

        :param obj:
        :param timeout:
        :return:
        """
        from palp.conn import redis_conn

        redis_conn.zadd(self.redis_key, {obj.to_json(): obj.priority})

    def get(self, timeout=None):
        """
        获取任务（这里是返回的对象）

        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.bzpopmin(self.redis_key, timeout=timeout)
        if result:
            return LoadRequest.load_from_json(result[1].decode())

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.zcard(self.redis_key) == 0



"""
    shell 交互生成文件

    测试 python shell.py xxx
"""
import sys
import argparse
from pathlib import Path
from palp.tool.create_spider import CreateSpider
from palp.tool.create_project import CreateProject


def generator():
    if len(sys.argv) == 1:
        sys.argv.append('--h')  # 无参数默认输出帮助

    parser = argparse.ArgumentParser(description='Palp 创建爬虫模板')
    parser.add_argument('-p', '--project', help='创建爬虫项目如：palp create -p <project_name>')
    parser.add_argument('-s', '--spider', nargs='+', help='创建爬虫如：palp create -s <spider_name> <spider_type>')
    args = parser.parse_args()

    # 创建 spider
    if args.spider:
        args_spider = args.spider
        if len(args_spider) == 1:
            spider_name = args_spider[0]
            spider_type = 1
        else:
            spider_name = args_spider[0]
            spider_type = int(args_spider[1])

        if spider_type not in [1, 2, 3, 4]:
            raise ValueError('spider_type 不在允许范围内！')
        CreateSpider(spider_name, spider_type).create()

    # 创建项目
    if args.project:
        project_name = args.project
        CreateProject(project_name).create()


def helper():
    version_path = Path(__file__).parent.parent.joinpath('VERSION')
    with open(version_path, 'r', encoding='utf-8') as f:
        version = f.read()

    print(f'palp {version}\n\nPalp 操作命令如下')
    print('\tpalp create [options] [args]')
    print('\n可选 options:')
    cmds = {'-p': '即 project，创建爬虫项目', '-s': '即 spider，创建爬虫'}
    for cmdname, cmdclass in sorted(cmds.items()):
        print('\t%s\t\t%s' % (cmdname, cmdclass))

    print('\n可选 args:')
    cmds = {
        '1': '创建 spider 时，创建普通 spider',
        '2': '创建 spider 时，创建分布式 spider',
        '3': '创建 spider 时，创建周期 spider',
        '4': '创建 spider 时，创建跳转 spider（辅助其余 spider）'
    }
    for cmdname, cmdclass in sorted(cmds.items()):
        print('\t%s\t\t%s' % (cmdname, cmdclass))

    print('\n示例:')
    print('\tpalp create -p demo')
    print('\tpalp create -s baidu 1')


def main():
    args = sys.argv
    if len(args) < 2:
        helper()
        return

    command = args.pop(1)
    if command == 'create':
        generator()
    elif command == '-c':
        generator()
    else:
        helper()


if __name__ == '__main__':
    helper()



from palp.network.request import Request
from palp.network.request_method import RequestGet, RequestPost, RequestDelete, RequestOptions, RequestHead, \
    RequestPatch
from palp.network.downloader import ResponseDownloader
from palp.network.downloader_httpx import ResponseDownloaderByHttpx
from palp.network.downloader_requests import ResponseDownloaderByRequests
from palp.network.response import Response
from palp.network.response_httpx import HttpxResponse
from palp.network.response_requests import RequestsResponse



"""
    redis set 过滤

    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
"""

from palp import settings
from quickdb import RedisLock
from palp.network.request import Request
from palp.filter.filter import FilterBase


class RedisSetFilter(FilterBase):

    def is_repeat(self, obj, **kwargs):
        """
        获取对应的指纹，通过 redis 的 set 去重

        :param obj:
        :param kwargs:
        :return:
        """
        from palp.conn import redis_conn

        fingerprint = self.fingerprint(obj=obj)

        if isinstance(obj, Request):
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_REQUEST
        else:
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_ITEM

        if settings.STRICT_FILTER:
            with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'Request'):
                return self.judge(redis_key_filter, fingerprint)
        else:
            return self.judge(redis_key_filter, fingerprint)

    def judge(self, f, fingerprint):
        """
        进行判断

        :param f: 判断条件或方法之类
        :param fingerprint: 指纹
        :return:
        """
        from palp.conn import redis_conn

        if redis_conn.sismember(f, fingerprint):
            return True

        redis_conn.sadd(f, fingerprint)

        return False
